/*
 ATtiny GPS LongLat Logger
 Version: 0.1
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 9/10/2013
 Last Modified: 9/10/2013
 
 */
 
#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

#ifndef boolean
	typedef uint8_t boolean;
#endif
#ifndef bool
	typedef uint8_t bool;
#endif 
#ifndef byte
	typedef uint8_t byte;
#endif

#ifndef NULL
#define NULL ((void *)0)
#endif

#define LOW 0
#define HIGH 1
#define false 0
#define true 1

#define TWI_SDA_PIN PB4
#define TWI_SCL_PIN PB3

#define I2C_DELAY_USEC 4
#define I2C_READ 1
#define I2C_WRITE 0
#define NOREPEATSTART 0
#define REPEATSTART 1

// Initialize SCL/SDA pins and set the bus high
void SoftI2cMasterInit(void) {
	DDRB |= (1<<TWI_SDA_PIN);
	PORTB |= (1<<TWI_SDA_PIN);
	DDRB |= (1<<TWI_SCL_PIN);
	PORTB |= (1<<TWI_SCL_PIN);
}

// De-initialize SCL/SDA pins and set the bus low
void SoftI2cMasterDeInit(void) {
	PORTB &= ~(1<<TWI_SDA_PIN);
	DDRB &= ~(1<<TWI_SDA_PIN);
	PORTB &= ~(1<<TWI_SCL_PIN);
	DDRB &= ~(1<<TWI_SCL_PIN);
}

// Write a byte to I2C
bool SoftI2cMasterWrite(uint8_t data) {
	// Write byte
	for (uint8_t m = 0x80; m != 0; m >>= 1) {
		// Don't change this loop unless you verify the change with a scope
		if (m & data) { 
			PORTB |= (1<<TWI_SDA_PIN); 
		}
		else { 
			PORTB &= ~(1<<TWI_SDA_PIN); 
		}
		PORTB |= (1<<TWI_SCL_PIN);
		_delay_us(I2C_DELAY_USEC);
		PORTB &= ~(1<<TWI_SCL_PIN);
	}
	// get Ack or Nak
	DDRB &= ~(1<<TWI_SDA_PIN);
	// Enable pullup
	PORTB |= (1<<TWI_SDA_PIN);
	PORTB |= (1<<TWI_SCL_PIN);
	uint8_t rtn = bit_is_set(PINB, TWI_SDA_PIN);
	PORTB &= ~(1<<TWI_SCL_PIN);
	DDRB |= (1<<TWI_SDA_PIN);
	PORTB &= ~(1<<TWI_SDA_PIN);
	return rtn == 0;
}

// Issue a start condition
bool SoftI2cMasterStart(uint8_t addressRW) {
	PORTB &= ~(1<<TWI_SDA_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTB &= ~(1<<TWI_SCL_PIN);
	return SoftI2cMasterWrite(addressRW);
}

// Issue a restart condition
bool SoftI2cMasterRestart(uint8_t addressRW) {
	PORTB |= (1<<TWI_SDA_PIN);
	PORTB |= (1<<TWI_SCL_PIN);
	_delay_us(I2C_DELAY_USEC);
	return SoftI2cMasterStart(addressRW);
}

// Issue a stop condition
void SoftI2cMasterStop(void) {
	PORTB &= ~(1<<TWI_SDA_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTB |= (1<<TWI_SCL_PIN);
	_delay_us(I2C_DELAY_USEC);
	PORTB |= (1<<TWI_SDA_PIN);
	_delay_us(I2C_DELAY_USEC);
}

// Issue the device start and the address to write
bool SoftI2cMasterAddress(uint8_t deviceAddr, uint16_t addressRW, bool repeatedStart) {
	if (!SoftI2cMasterStart((deviceAddr<<1) | I2C_WRITE)) return false;
	if (!SoftI2cMasterWrite((addressRW >> 8))) return false; // MSB
	if (!SoftI2cMasterWrite((addressRW & 0xFF))) return false; // LSB
	if (repeatedStart) { if (!SoftI2cMasterRestart((deviceAddr<<1) | I2C_READ)) return false; }
	
	return true;
}

// Write 1 byte to the EEPROM
bool soft_i2c_eeprom_write_byte(uint8_t deviceAddr, uint16_t writeAddress, byte writeByte) {
	
	// Issue start condition and send device address
	SoftI2cMasterAddress(deviceAddr, writeAddress, NOREPEATSTART);
	
	// Write the byte
	if (!SoftI2cMasterWrite(writeByte)) return false;
	
	// Issue a stop condition
	SoftI2cMasterStop();
	
	// Delay 10ms for the write to complete, depends on the EEPROM you use
	_delay_ms(10);
	
	return true;
}

// Used from http://interface.khm.de/index.php/lab/experiments/sleep_watchdog_battery/
void system_sleep(void) {
	set_sleep_mode(SLEEP_MODE_PWR_DOWN); // Set sleep mode
	sleep_mode(); // System sleeps here
}

// Setup the ATtiny
void setup(void) {
	UBRRL = 12; // 9,600 Baud rate
	sbi(UCSRA, U2X); // Double rate
	sbi(UCSRB, RXEN); // Receiver enable
	sbi(UCSRB, TXEN); // Transmitter enable
	
	SoftI2cMasterInit(); // Initialise software I2C/TWI
	
	sei(); // Turn on interrupts
}
